#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Dec  9 12:48:26 2020

@author: jacob
"""
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt
        
def hm(plat, data_path):
    plate_name = plat.get_plate_name()
    params = plat.get_all_params()
    
    # List for making heatmaps
    row_letters = ["A", "B", "C", "D", "E", "F", "G", "H"]
    cols = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12]

    w, h = 3, 96
    hm_data_gr = [[0 for x in range(w)] for y in range(h)]
    hm_data_ymax = [[0 for x in range(w)] for y in range(h)]
    
    index = 0
    for i in range(8):
        for j in range(12):
            no_growth_flag=False
            well = row_letters[i] + str(cols[j])
            hm_data_gr[index][0] = row_letters[i]
            hm_data_gr[index][1] = cols[j]
            
            hm_data_ymax[index][0] = row_letters[i]
            hm_data_ymax[index][1] = cols[j]
            
            try:
                hm_data_ymax[index][2] = params[well][1]
                if params[well][1] < 0.2:
                    no_growth_flag=True
            except KeyError:
                hm_data_ymax[index][2] = None
            
            if no_growth_flag:
                hm_data_gr[index][2] = 0
            else:
                try:
                    hm_data_gr[index][2] = params[well][0]
                except KeyError:
                    hm_data_gr[index][2] = None      
            index += 1
            
    # Tranposes data
    hm_df_gr = pd.DataFrame(hm_data_gr)
    hm_df_gr.columns = ["Rows", "Columns", "GR"]
    hm_df_gr = hm_df_gr.pivot(index="Rows", columns="Columns", values="GR")
    
    hm_df_ymax = pd.DataFrame(hm_data_ymax)
    hm_df_ymax.columns = ["Rows", "Columns", "Ymax"]
    hm_df_ymax = hm_df_ymax.pivot(index="Rows", columns="Columns", values="Ymax")

    # Formatting heatmap to align with 96 well plate
    sns.set(font_scale=3)
    f, ax = plt.subplots(figsize=(42,28))
    with sns.axes_style("white"):
        sns.heatmap(hm_df_gr, ax=ax, mask=hm_df_gr.isnull(), linewidth=0.5, cmap="magma", annot=True, vmin=0.05)
    ax.set_title(plate_name + ": Raw Growth Rate Values\n\n")
    plt.yticks(rotation=0)
    ax.xaxis.tick_top()  
    ax.set_xlabel('')
    ax.set_ylabel('')
    
    plt.savefig(data_path + "Heatmaps/GR/" + plate_name) 
    plt.close()
    
     # Formatting heatmap to align with 96 well plate
    sns.set(font_scale=3)
    f, ax = plt.subplots(figsize=(42,28))
    with sns.axes_style("white"):
        sns.heatmap(hm_df_ymax, ax=ax, mask=hm_df_ymax.isnull(), linewidth=0.5, cmap="magma", annot=True, vmin=0.2)
    ax.set_title(plate_name + ": Raw Ymax Values\n\n")
    plt.yticks(rotation=0)
    ax.xaxis.tick_top()  
    ax.set_xlabel('')
    ax.set_ylabel('')

    plt.savefig(data_path + "Heatmaps/Ymax/" + plate_name)
    plt.close()
    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Jul 30 13:10:02 2020

Updated on June 2, 2021

@author: jacob
"""

""" Extracts and cleans the relevant data from the standard output of the Epoch2
    Biotek machine in the 30C room of Striesinger
    Hall, University of Oregon """

import pandas as pd
import os
from sys import exit

def extr(DATA_PATH):
    print("Extracting and cleaning...\n")
    data_file = ""
    # Data files from the experiment
    try:
        n_files = 0
        for file in os.listdir(DATA_PATH + "/Data"):
            # if the element is an xlsx file then
            if file[-5:] == ".xlsx" and file[0] != "~":
                data_file = file
                n_files += 1
                if n_files > 1:
                    print("Too many data files, combine into one excel file. Exiting...")
                    exit(1)
    except FileNotFoundError:
        print("Could not find file or directory " + DATA_PATH + "/Data - exiting...")
        exit(0)
        
    # Create directory for parsed data
    try:
        os.mkdir(DATA_PATH + "Data/Raw_OD")
    except OSError:
        pass
    
    # Checks to see if experimental info exists is present
    info_file = ""
    try:
        for file in os.listdir(DATA_PATH):
            # if the element is an xlsx file then
            if file == "experiment_info.txt" or file == "experimental_info.txt":
                info_file = file 
                # Pulling info from experimental_info.txt
                print("Using information from experiment_info.txt\n")
                break
    except FileNotFoundError:
        print("FileNotFoundError, try making an experiment_info.txt file and trying again")
        exit(0)
    if info_file == "":
        print("FileNotFoundError, try making an experiment_info.txt file and trying again")
        print("If this file already exists, try renaming it, deleting the last letter, then retyping the last letter")
        exit(0)
      
    # Experimental info
    ignore_plates = []
    plate_names = []
    num_measurements = ""
    water_plate_pres = False
    
    # Full spreadsheet
    xls = pd.ExcelFile(DATA_PATH + "Data/" + data_file)

    # List of all sheets in spreadsheet
    df_dict = pd.read_excel(xls, None)
        
    
    water_plates = ["top water", "bottom water", "water"]
    with open(DATA_PATH + info_file, "r") as info_f:
        for line in info_f:
            # Read in plate names
            if "plate names" in line.lower():
                line = info_f.readline()
                while line != "\n" and line != "":
                    # Checks if water plate
                    if line.lower().replace("\n", "") not in water_plates:
                        plate_names.append(line.replace("\n", ""))
                    else:
                        water_plate_pres = True
                    line = info_f.readline()
                    
            # Read in plates to ignore
            elif "plates to ignore" in line.lower():
                line = info_f.readline()
                while line != "\n" and line != "":
                    if line.lower().replace("\n", "") != "none":
                        ignore_plates.append(line.strip("\n "))
                    line = info_f.readline()
                
            # Read in the number of measurements
            elif "number of measurements" in  line.lower():
                line = info_f.readline()
                while line != "\n" and line != "":
                    try:
                        num_measurements = int(line.strip("\n "))
                    except ValueError:
                        print("Number of measurements should be an integer. Exiting...")
                        exit(0)
                    line = info_f.readline()
    plate_names = plate_names[::-1] 
                    
    # Experimental information has been loaded, now processing file 
    offset = 0
    df_dict = {}
    if water_plate_pres:
        offset = 1
    
    column_row = -1
    stop = len(plate_names)
    for i in range(0, stop):
        plate_name = plate_names[i]
        if plate_name in ignore_plates:
            continue
        else:
            sheet_name = "Plate " + str(i + 1 + offset) + " - Sheet1"
            # Format the dataframe to save correctly to an excel file
            df = pd.read_excel(xls, sheet_name)
            # Search through spreadsheet to find the start of the data
            if column_row < 0:
                for i in range(0, 40):
                    column_names = df.iloc[[i], 1:99].values[0]
                    #column_names = column_names[0]
                    if column_names[0] == "Time":
                        column_row = i
                        break
            # Didn't find data in the correct format
            if column_row < 0:
                print("Excel file is not in the correct format, see examples")
                exit(0)
                
            df_new = pd.DataFrame(df.iloc[column_row:(column_row + num_measurements + 1), 1:99])
            df_new.reset_index(drop=True, inplace=True)
            df_new.columns = column_names
            
            # Drop the temperature column
            df_new.drop(columns=["T° 600"], inplace=True)
            
            # Calculate the measurement times in minutes
            times = df_new.Time.values
            times_formatted = []
            minutes = 0
            for time in times:
                # Retrieve number days between first measurement and current
                try:
                    days = time.day
                except AttributeError:
                    days = 0
                # Retrieve hours
                try:
                    hours = time.hour
                except AttributeError:
                    hours = 0
                # Retrieve minutes
                try:
                    mins = time.minute
                except AttributeError:
                    mins = 0
                # Calculate minutes passed
                minutes = (days * 24 + hours) * 60 + mins
                total_hours = minutes / 60
                times_formatted.append(total_hours)
                
            # Assign newly formatted times to column
            df_new["Time"] = times_formatted
            df_new.drop(df.index[0], inplace=True)
            
            # Store dataframes in a dictionary
            df_new.name = plate_name
            df_dict[plate_name] = df_new
            
            # Write dataframes to excel
            file_path = DATA_PATH + "Data/Raw_OD/" + plate_name + "_OD600.xlsx"
            df_new.to_excel(file_path, index=False)
            
    return df_dict
                
        
if __name__ == "__main__":
    print("Enter name of screen directory (Ex. 'Screen1')")
    inp = input("- ")
    extr("../Screens/" + inp + "/")


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Dec  9 12:46:52 2020

Updated on June 3, 2021

@author: jacob
"""
import pandas as pd
import matplotlib.pyplot as plt
import scipy.optimize as optim
import numpy as np
from scipy import stats

#from heatmap import heatmap_gr, heatmap_ymax
#from graph_repl import graph_repls

# How many time points are graphed
XSCALE = 97

def graph_avg(df_dict, con_data, exp_data, con_name, exp_name, data_path, plate_list, hm_flag=False, log_flag=False):
    """ Plot Formatting """
    # You typically want your plot to be ~1.33x wider than tall.
    # Common sizes: (10, 7.5) and (12, 9)
    plt.figure(figsize=(10, 7.5))
    
    con_color = "#0466c8"
    exp_color = "#d62828"

    # Remove the plot frame lines. They are unnecessary
    ax = plt.subplot(111)
    ax.spines["top"].set_visible(False)
    ax.spines["bottom"].set_visible(False)
    ax.spines["right"].set_visible(False)
    ax.spines["left"].set_visible(False)
    
    # Set background to white
    ax.set_facecolor('white')

    # Ensure that the axis ticks only show up on the bottom and left of the plot.
    ax.get_xaxis().tick_bottom()
    ax.get_yaxis().tick_left()

    # Limit the range of the plot to only where the data is.
    plt.ylim(0, 2)
    plt.xlim(0, XSCALE)

    # Make sure your axis ticks are large enough to be easily read.
    plt.yticks(np.arange(0, 1.7, 0.2), [str(round(x, 1)) for x in np.arange(0, 1.7, 0.2)], fontsize=14)
    plt.xticks(np.arange(0, XSCALE, 24), [str(round(x,1)) for x in np.arange(0, XSCALE, 24)], fontsize=14)

    # Provide tick lines across the plot to help your viewers trace along the axis ticks.
    for y in np.arange(0, 1.7, 0.2):
        plt.plot(range(0, XSCALE), [y] * len(range(0, XSCALE)), "--", lw=0.5, color="black", alpha=0.3)
        
    """ Calculations """   
    # Parameter fits for individual control wells
    control_grs = []
    control_ymaxs = []
    # Storing wells to compute average of replicate
    control_wells = []
    con_avg = con_name + "_avg"
    
    # Lists of parameter values for the experimental replicate
    exp_grs = []
    exp_ymaxs = []
    # Storing wells to compute average of replicate
    exp_wells = []
    exp_avg = exp_name + "_avg"
    
    avg_df = pd.DataFrame()
    
    # Calculate parameter values for individual wells
    for plate_name in con_data.keys():
        # Plate
        df = df_dict[plate_name]
        plat = plate_list[plate_name]
        # Wells in that specific plate that belong to given control replicate
        wells = con_data[plate_name]
        for well in wells:
            if well == "":
                break
            control_wells.append(df[well])
            gr, ymax, line = fit_model(df, well)
            plat.add_params(gr, ymax, well)
            if gr < 2:
                control_grs.append(gr)
            if ymax < 2: 
                control_ymaxs.append(ymax)
                
    for plate_name in exp_data.keys():
        # Plate
        df = df_dict[plate_name]
        plat = plate_list[plate_name]
        # Wells in that specific plate that belong to given control replicate
        wells = exp_data[plate_name]
        for well in wells:
            if well == "":
                break
            exp_wells.append(df[well])
            gr, ymax, line = fit_model(df, well)
            plat.add_params(gr, ymax, well)
            if gr < 2:
                exp_grs.append(gr)
            if ymax < 2: 
                exp_ymaxs.append(ymax)
                
    avg_df["Time"] = df["Time"]
    # Calculate averages for replicates
    con_mean, con_std, con_ci = avg_well(control_wells)
    avg_df[con_avg] = con_mean
    avg_df[con_name + "_std"] = con_std
    avg_df[con_name + "_ci"] = con_ci
    
    exp_mean, exp_std, exp_ci = avg_well(exp_wells)
    avg_df[exp_avg] = exp_mean
    avg_df[exp_name + "_std"] = exp_std
    avg_df[exp_name + "_ci"] = exp_ci
    
    # Parameter fits for average control model
    con_gr, con_ymax, con_line = fit_model(avg_df, con_avg)
    # Parameter fits for average exp model
    exp_gr, exp_ymax, exp_line = fit_model(avg_df, exp_avg)
    
    # T-test for growth rate and ymax parameter values
    gr_stats = t_test(exp_grs, control_grs)
    ymax_stats = t_test(exp_ymaxs, control_ymaxs)
    # P-values
    gr_pval = gr_stats[1]
    ymax_pval = ymax_stats[1]
    
    if con_ymax > 0.01 and exp_ymax > 0.01:
        # Normalize experimental parameters with control parameters
        gr_ratio = (exp_gr / con_gr) 
        ymax_ratio = (exp_ymax / con_ymax)
    else:
        gr_ratio = 0
        ymax_ratio = 0
            
        # Symbols on graph to indicate better growth by experimental strain
    better_gr = ""
    if gr_ratio > 1:
        better_gr += "^ "
    better_ymax = ""
    if ymax_ratio > 1:
        better_ymax += "^ "
    
    """ Graphing """
    # Graph average experimental line
    plt.plot(avg_df["Time"], avg_df[exp_avg], color=exp_color, label=(exp_name), linewidth=3.0)
    # plt.plot(*exp_line, 'r', linestyle = "--", color=exp_color, linewidth=1)
    # Confidence intervals
    exp_ci_hi =  avg_df[exp_avg] + avg_df[exp_name + "_ci"]
    exp_ci_low = avg_df[exp_avg] - avg_df[exp_name + "_ci"]
    plt.plot(avg_df["Time"], exp_ci_hi, color=exp_color, linestyle=":", linewidth=1.5)
    plt.plot(avg_df["Time"], exp_ci_low, color=exp_color, linestyle=":", linewidth=1.5)
        
    # Graph average control line  
    plt.plot(avg_df["Time"], avg_df[con_avg], color=con_color, label=(con_name), linewidth=3.0)
    # plt.plot(*con_line, 'r', linestyle = "--", color=con_color, linewidth=1)
    # Confidence intervals
    con_ci_hi =  avg_df[con_avg] + avg_df[con_name + "_ci"]
    con_ci_low = avg_df[con_avg] - avg_df[con_name + "_ci"]
    plt.plot(avg_df["Time"], con_ci_hi, color=con_color, linestyle=":", linewidth=1.5)
    plt.plot(avg_df["Time"], con_ci_low, color=con_color, linestyle=":", linewidth=1.5)
    
    # Plot histograms
    # graph_repls(con_grs, con_ymaxs, exp_grs, exp_ymax,con_name, exp_name, data_path)
    
    # Place a legend to the right
    lgd = ax.legend(
                    loc = 'upper right', 
                    borderaxespad = 0., 
                    facecolor = 'white', 
                    fontsize = 16)
    
    # Format P-values
    if gr_pval < 0.001:
        gr_pval = "<0.001"
    else:
        gr_pval = round(gr_pval, 3)
    if ymax_pval < 0.001:
        ymax_pval = "<0.001"
    else:
        ymax_pval = round(ymax_pval, 3)
        
    plt.title(f"{exp_name} vs. {con_name}- GR ratio:{round(gr_ratio, 3)} ({gr_pval}) Ymax ratio: {round(ymax_ratio, 3)} ({ymax_pval})", fontsize=20)
    path =  data_path + "Graphs/Averages/" + exp_name + "_vs_" + con_name
    plt.savefig(path, bbox_extra_artists=(lgd,), bbox_inches='tight')
    plt.close()
    
# Graph each well of a replicate
def graph_indiv(df_dict, repl_data, repl_name, data_path, plate_list, log_flag=False):
    """ Graph Formatting """
    # You typically want your plot to be ~1.33x wider than tall.
    # Common sizes: (10, 7.5) and (12, 9)
    plt.figure(figsize=(10, 7.5))

    # Remove the plot frame lines. They are unnecessary
    ax = plt.subplot(111)
    ax.spines["top"].set_visible(False)
    ax.spines["bottom"].set_visible(False)
    ax.spines["right"].set_visible(False)
    ax.spines["left"].set_visible(False)
    
    # Set background to white
    ax.set_facecolor('white')

    # Ensure that the axis ticks only show up on the bottom and left of the plot.
    ax.get_xaxis().tick_bottom()
    ax.get_yaxis().tick_left()

    # Limit the range of the plot to only where the data is.
    plt.ylim(0, 2)
    plt.xlim(0, XSCALE)

    # Make sure your axis ticks are large enough to be easily read.
    plt.yticks(np.arange(0, 1.7, 0.2), [str(round(x, 1)) for x in np.arange(0, 1.7, 0.2)], fontsize=14)
    plt.xticks(np.arange(0, XSCALE, 24), [str(round(x,1)) for x in np.arange(0, XSCALE, 24)], fontsize=14)

    # Provide tick lines across the plot to help your viewers trace along the axis ticks.
    for y in np.arange(0, 1.7, 0.2):
        plt.plot(range(0, XSCALE), [y] * len(range(0, XSCALE)), "--", lw=0.5, color="black", alpha=0.3)

    # Graph each replicate well
    n = 0
    for plate_name in repl_data.keys():
        # Plate
        df = df_dict[plate_name]
        plat = plate_list[plate_name]
        # Wells in that specific plate that belong to given control replicate
        wells = repl_data[plate_name]
        # Counter for number of viable wells
        n = 0
        
        for well in wells:
            if well == "":
                break
            try:
                gr, ymax = plat.get_params(well)
            except KeyError:
                gr, ymax, line = fit_model(df, well)
                plat.add_params(gr, ymax, well)
            
            if ymax > 0.2:
                n += 1
            plt.plot(df["Time"], df[well], label=well, linewidth=2.5)
                
    # Place a legend to the right
    lgd = ax.legend(
                    loc = 'upper right', 
                    borderaxespad = 0., 
                    facecolor = 'white', 
                    ncol = 2,
                    fontsize = 16)
        
    plt.title(f"{repl_name} Isolates: ({n} isolates with growth)", fontsize=24)
    path =  data_path + "Graphs/" + repl_name 
    plt.savefig(path, bbox_extra_artists=(lgd,), bbox_inches='tight')
    plt.close()
    
""" Auxillary Functions """
# P-test on individual wells
def t_test(data1, data2) -> int:
    ind_ttest = stats.ttest_ind(data1, data2)
    return ind_ttest

def fit_model(df, well):
    # Calculate exponential portion of data for line fit
        exp = exponential_section(df, well)
        # Fitting lines to exponential portions to graph
        slope = 0
        if not exp.empty:
            slope, line = fit_line(exp, well)
        else:
            line = [(0,0), (0,0)]

        # Fit a logistical model to calculate growth rate
        p0 = np.random.exponential(size=3) # Initialize random values
        bounds = (0, [10000000., 100., 10000000.]) # Set bounds
        # Prepare model 
        xt = np.array(df["Time"])
        yt = np.array(df[well])
        
        # If no logistic curve can be fit, default to less sophisticated method of fitting line to exponentional section of the graph
        try:
            # Fit model 1
            (a, gr, ymax), cov = optim.curve_fit(logistic, xt, yt, bounds=bounds, p0=p0)
        except (RuntimeError, ValueError):
            gr = slope
            ymax = max(df[well])
        return gr, ymax, line

# Estimates expinential growth section of growth curve to compute growth rate
def exponential_section(df, well):
    ymax = max(df[well])
    ymin = min(df[well])
    ymid = (ymax + ymin) / 2.0
    span = ymax - ymin
    low = ymid - (span * 0.40)
    high = ymid + (span * 0.40)
    exp = df.loc[(df[well] >= low) & (df[well] <= high)]
    return exp[["Time", well]]

# Fits a line to a given section of a graph. Returns the slope and endpoints of the line
def fit_line(exp, well):
    exp["Time"] = pd.to_numeric(exp["Time"])
    exp[well] = pd.to_numeric(exp[well])
    slope, intercept, r_value, p_value, std_err = stats.linregress(exp["Time"], exp[well])
    x1 = int(exp.iloc[:1,:1].values)
    x2 = int(exp.iloc[-1:,:1].values)
    y1 = x1 * slope + intercept
    y2 = x2 * slope + intercept
    p1 = (x1, x2)
    p2 = (y1, y2)
    line = [p1, p2]
    return slope, line

# Logistical funtion used to model growth rate
def logistic(t, a, b, c):
    return c / (1 + a * np.exp(-b*t))

# Returns several statistics for group of dataframe columns (mean, SD, 95% CI)
def avg_well(well_list):   
    col = pd.DataFrame()
    n = 0
    for i in range(len(well_list)):
        well = well_list[i]
        # Check if well was viable
        if max(well) > 0.2:
            col[str(i)] = well
            n += 1 
    mean = col.mean(axis=1).reset_index(drop=True)
    std = col.std(axis=1).reset_index(drop=True)
    std_err = col.sem(axis=1).reset_index(drop=True)
    ci95 = std_err * stats.t.ppf((1 + 0.95) / 2, n - 1)
    return mean, std, ci95  


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Jun  3 17:55:52 2021

@author: jacob
"""
"""
Object to hold data about each 96 well plate in an experiment,
including a mapping of where and how replicates are grouped """
class Plate(object):
    def __init__(self, name):
        self.name = name
        self.replicate_map = {}
        self.parameter_map = {}
        
    def __repl__(self):
        rep = f"{self.name}:\n"
        for repl_name in self.replicate_map.keys():
            rep += f"{repl_name}: {self.replicate_map[repl_name]}"
        return rep
        
    def __str__(self):
        rep = f"{self.name}:\n"
        for repl_name in self.replicate_map.keys():
            rep += f"{repl_name}: {self.replicate_map[repl_name]}\n"
        return rep
        
    def add_repl(self, repl_name, wells):
        self.replicate_map[repl_name] = wells
        
    # Keeps track of growth rates and ymaxs in each well of the plate
    # Note: not safe, wrap in try/catch statement (KeyError)
    def add_params(self, gr, ymax, well):
        self.parameter_map[well] = [gr, ymax]
     
    # Note: not safe, wrap in try/catch statement (KeyError)
    def get_params(self, well):
        return self.parameter_map[well][0], self.parameter_map[well][1]

    
    def get_all_params(self):
        return self.parameter_map
        
    def get_wells(self, repl_name):
        return self.replicate_map[repl_name]
    
    def get_repl_names(self):
        return list(self.replicate_map.keys())
    
    def get_plate_name(self):
        return self.name
    
    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Dec 10 09:31:00 2020

@author: jacob
"""
import pandas as pd
import matplotlib.pyplot as plt
import scipy.optimize as optim
import numpy as np
from scipy import stats
import seaborn as sns

# Graph the average of all the wells in a given replicate
def graph_repls(df, repl_wells, repl_n, data_path, xscale):
    file_n = df.name
    # Parameter fits
    control_grs = []
    control_ymax = []
    exp_grs = []
    exp_ymax = []
    
    for well in repl_wells:
        # Get parameters from control well
        cwell = well + "_No_PO"
        gr1, ymax1, line1 = fit_model(df, cwell)
        # Remove erroneous values
        if gr1 < 1.5:
            control_grs.append(gr1)
        if ymax1 < 2:
            control_ymax.append(ymax1)
        # Get parameters from experimental well
        ewell = well + "_PO"
        gr2, ymax2, line2 = fit_model(df, ewell)
        if gr2 < 1.5:
            exp_grs.append(gr2)
        if ymax2 < 2:
            exp_ymax.append(ymax2)
        
    # Test the significance of the fitted parameters
    gr_stats = t_test(exp_grs, control_grs)
    ymax_stats = t_test(exp_ymax, control_ymax)
    # Get P-values from statistics
    gr_pval = gr_stats[1]
    ymax_pval = ymax_stats[1]
            
    # Calculate average of wells to graph
    exp_avg, exp_std, exp_ci95 = avg_well(df, repl_wells, control=False)
    control_avg, control_std, control_ci95 = avg_well(df, repl_wells, control=True)
    
    df["Exp_avg_" + str(repl_n)] = exp_avg
    df["Control_avg_" + str(repl_n)] = control_avg
    
    gr1, ymax1, line1 = fit_model(df, "Control_avg_" + str(repl_n))
    gr2, ymax2, line2 = fit_model(df, "Exp_avg_" + str(repl_n))
    
    # Convert the list of param values to dataframe for graphing
    gr_df = pd.DataFrame(control_grs, columns=["GR"])
    gr_df["Strain"] = "Control"
    gr_df_inter = pd.DataFrame(exp_grs, columns=["GR"])
    gr_df_inter["Strain"] = "Experimental"
    
    ymax_df = pd.DataFrame(control_ymax, columns=["Ymax"])
    ymax_df["Strain"] = "Control"
    ymax_df_inter = pd.DataFrame(exp_ymax, columns=["Ymax"])
    ymax_df_inter["Strain"] = "Experimental"
    
    gr_df = pd.concat([gr_df, gr_df_inter], ignore_index=True)
    ymax_df = pd.concat([ymax_df, ymax_df_inter], ignore_index=True)
    
    if ymax1 > 0.2 and ymax2 > 0.2:
        gr_ratio = (gr2 / gr1) 
        ymax_ratio = (ymax2 / ymax1)
    else:
        gr_ratio = 0
        ymax_ratio = 0
            
    # Includes significance in graph
    gr_sig = ""
    if gr_pval <= 0.001:
        gr_sig = " GR:*** "
    elif gr_pval <= 0.01:
        gr_sig = " GR:**  "
    elif gr_pval <= 0.05:
        gr_sig = " GR:*   "
    else:
        gr_sig = " GR: NS "
            
    ymax_sig = ""
    if ymax_pval <= 0.001:
        ymax_sig = " Ymax:*** "
    elif ymax_pval <= 0.01:
        ymax_sig = " Ymax:**  "
    elif ymax_pval <= 0.05:
        ymax_sig = " Ymax:*   "
    else:
        ymax_sig = " Ymax: NS "

    """ Plotting """
    control_color = '#2980b9'
    exp_color = '#e74c3c'
    
    # You typically want your plot to be ~1.33x wider than tall.
    fig = plt.figure(figsize=(8,8))
    fig.tight_layout(pad=4.0)
    ax = []
    # Image 1: Growth curves
    # Remove the plot frame lines. They are unnecessary
    ax0 = fig.add_subplot(211)
    ax.append(ax0)
    ax[0].spines["top"].set_visible(False)
    ax[0].spines["bottom"].set_visible(False)
    ax[0].spines["right"].set_visible(False)
    ax[0].spines["left"].set_visible(False)
    
    # Set background to white
    ax[0].set_facecolor('white')

    # Ensure that the axis ticks only show up on the bottom and left of the plot.
    ax[0].get_xaxis().tick_bottom()
    ax[0].get_yaxis().tick_left()

    # Limit the range of the plot to only where the data is.
    ax[0].set_ylim(0, 2)
    ax[0].set_xlim(0, xscale)

    # Make sure your axis ticks are large enough to be easily read.
    ax[0].set_yticks(np.arange(0, 1.7, 0.2))
    ax[0].set_yticklabels([str(round(x, 1)) for x in np.arange(0, 1.7, 0.2)], fontsize=14)
    ax[0].set_xticks(np.arange(0, xscale, 12))
    ax[0].set_xticklabels([str(round(x,1)) for x in np.arange(0, xscale, 12)], fontsize=12)
    
    # Provide tick lines across the plot to help your viewers trace along the axis ticks.
    for y in np.arange(0, 1.7, 0.2):
        ax[0].plot(range(0, xscale), [y] * len(range(0, xscale)), "--", lw=0.5, color="black")
    
    ax[0].set_xlabel("Time (h)", fontsize=12)
    ax[0].set_ylabel("OD 600", fontsize=14)
    
    # Graph Averags
    ax[0].plot(df["Time"], control_avg, color=control_color,  label=("Control"), linewidth=2.0)
    ax[0].plot(*line1, 'b', linestyle = "--", linewidth=1)
    ax[0].plot(df["Time"], exp_avg, color=exp_color, label=("Experimental"), linewidth=2.0)
    ax[0].plot(*line2, 'r', linestyle = "--", linewidth=1)
    
    # Calculate 95% confidence intervals
    ciA_hi =  control_avg + control_ci95
    ciA_low = control_avg - control_ci95
    ciB_hi = exp_avg + exp_ci95
    ciB_low = exp_avg - exp_ci95
    # Graph 95% confidence intervals
    ax[0].plot(df["Time"], ciA_hi, color=control_color, linestyle=":", linewidth=1)
    ax[0].plot(df["Time"], ciA_low, color=control_color, linestyle=":", linewidth=1)
    ax[0].plot(df["Time"], ciB_hi, color=exp_color, linestyle=":", linewidth=1)
    ax[0].plot(df["Time"], ciB_low, color=exp_color, linestyle=":", linewidth=1)   
    
    # Place a legend to the right
    lgd = ax[0].legend(
                    borderaxespad = 0., 
                    facecolor = 'white', 
                    fontsize = 16)
    
    # Include a title  
    titl = f"Replicate {str(repl_n)}: {gr_sig} GR_ratio: {round(gr_ratio, 3)} {ymax_sig} Ymax_ratio: {round(ymax_ratio, 3)}"
    ax[0].set_title(titl, fontsize=24)
    
    # Image 2: histogram of growth rates
    ax1 = fig.add_subplot(223)
    ax.append(ax1)
    ax[1].set_xlabel("Growth rate value bins", fontsize=14)
    ax[1].set_ylabel("Frequency", fontsize=14)    
    ax[1].set_facecolor('white')

    # Ensure that the axis ticks only show up on the bottom and left of the plot.
    ax[1].get_xaxis().tick_bottom()
    ax[1].get_yaxis().tick_left()
    
    # Add vertical mean line to histogram
    control_mean = mean(control_grs)
    exp_mean = mean(exp_grs)
    ax[1].axvline(control_mean, color=control_color)
    ax[1].axvline(exp_mean, color=exp_color)
    
    # Plot histogram and estimated PDF
    sns.histplot(data=gr_df, x="GR", hue="Strain", 
                 ax=ax[1], legend=False, color=control_color, 
                 bins = 15, kde=False, multiple="layer",
                 palette=['#3867d6', '#eb3b5a'])  
    
    # Image 3: historgram of ymaxs
    ax2 = fig.add_subplot(224, sharey=ax[1])
    ax.append(ax2)
    ax[2].set_xlabel("Ymax value bins", fontsize=14)
    ax[2].set_ylabel(" ")
    ax[2].set_facecolor('white')

    # Ensure that the axis ticks only show up on the bottom and left of the plot.
    ax[2].get_xaxis().tick_bottom()
    ax[2].get_yaxis().tick_left()
    
    # Add vertical mean line to histogram
    control_mean = mean(control_ymax)
    exp_mean = mean(exp_ymax)
    ax[2].axvline(control_mean, color=control_color)
    ax[2].axvline(exp_mean, color=exp_color)
    
    # Plot histogram and estimated PDF
    sns.histplot(data=ymax_df, x = "Ymax", 
                 hue="Strain", ax=ax[2], legend=False,
                 color=control_color, bins = 15, 
                 kde=False, multiple = "layer",
                 palette=['#3867d6', '#eb3b5a'])  
    
    # Save the images  
    path = data_path + "Graphs/Replicates/" + file_n + "/avg_repl_" + str(repl_n)
    plt.savefig(path, bbox_extra_artists=(lgd,),bbox_inches='tight')
    plt.close()
    
""" Auxillary Functions """       
# P-test on individual wells
def t_test(data1, data2) -> int:
    ind_ttest = stats.ttest_ind(data1, data2)
    return ind_ttest

def mean(lst):
    return round((sum(lst) / len(lst)), 3)
    
def avg_well(df, col_list, control=False): 
    col_lst = []
    post = ""
    # Control wells
    if control:
        post = "_No_PO"
    # Exp wells
    else:
        post = "_PO"
        
    for well in col_list:
        col_lst.append(well + post)
        
    col = df.loc[:, col_lst]
    mean = col.mean(axis=1)
    std = col.std(axis=1)
    n = len(col_lst)
    std_err = col.sem(axis=1)
    ci95 = std_err * stats.t.ppf((1 + 0.95) / 2, n - 1)
    return mean, std, ci95

# Estimates expinential growth section of growth curve to compute growth rate
def exponential_section(df, well):
    ymax = max(df[well])
    ymin = min(df[well])
    ymid = (ymax + ymin) / 2.0
    span = ymax - ymin
    low = ymid - (span * 0.40)
    high = ymid + (span * 0.40)
    exp = df.loc[(df[well] >= low) & (df[well] <= high)]
    return exp[["Time", well]]

# Fits a line to a given section of a graph. Returns the slope and endpoints of the line
def fit_line(exp, well):
    slope, intercept, r_value, p_value, std_err = stats.linregress(exp["Time"], exp[well])
    x1 = int(exp.iloc[:1,:1].values)
    x2 = int(exp.iloc[-1:,:1].values)
    y1 = x1 * slope + intercept
    y2 = x2 * slope + intercept
    p1 = (x1, x2)
    p2 = (y1, y2)
    line = [p1, p2]
    return slope, line

def fit_model(df, well):
    # Calculate exponential portion of data for line fit
        exp = exponential_section(df, well)
        # Fitting lines to exponential portions to graph
        slope = 0
        if not exp.empty:
            slope, line = fit_line(exp, well)
        else:
            line = [(0,0), (0,0)]
                
        # Fit a logistical model to calculate growth rate
        p0 = np.random.exponential(size=3) # Initialize random values
        bounds = (0, [10000000., 100., 10000000.]) #Set bounds
        # Prepare model 1
        xt = np.array(df["Time"])
        yt = np.array(df[well])
            
        # If no logistic curve can be fit, default to less sophisticated method of fitting line to exponentional section of the graph
        try:
            # Fit model 1
            (a, gr, ymax), cov = optim.curve_fit(logistic, xt, yt, bounds=bounds, p0=p0)
        except (RuntimeError, ValueError):
            gr = slope 
            ymax = max(df[well])
        return gr, ymax, line

# Logistical funtion used to model growth rate
def logistic(t, a, b, c):
    return c / (1 + a * np.exp(-b*t))


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Aug 26 16:32:51 2019

Updated June 4, 2021

@author: jacob
"""

""" Program to automate the production of graphs, heatmaps, and significance
    tests (ttest) on growth curve data. Raw data should be named in the format
    protein_stressAndPlate_rawOD (pus6_31B_rawOD.xlsx) and put into a folder 
    called RawOD before being run through combineOD.py before being run
    through this program. 
        - direct well to well comparison between two plates
"""

import os
from sys import exit

# Helper files
if __package__ is None or __package__ == '':
    # uses current directory visibility
    from extract import extr
    from grph import graph_avg, graph_indiv
    from plate import Plate
    from heatmap import hm
else:
    # uses current package visibility
    from wellcompare.extract import extr
    from wellcompare.grph import graph_avg, graph_indiv
    from wellcompare.plate import Plate
    from wellcompare.heatmap import hm


# Enter folder name hard coded here if running graph.py as main for testing
DATA_PATH_HC = "../Screens/MTest"  
              
def proces(dp, hm_flag, log_flag):
    global DATA_PATH
    if dp[-1] != "/":
        DATA_PATH = dp + "/"
    else:
        DATA_PATH = dp
    
    # Extracts the data from the format output by the Epoch2 plate reader
    df_dict = extr(DATA_PATH)
            
    print("Graphing (this may take a minute)...\n")
    
    # Create Directories
    make_dir(DATA_PATH + "Graphs")
    make_dir(DATA_PATH + "Graphs/Averages")

    # Heatmap functionality is being updated
    make_dir(DATA_PATH + "Heatmaps")
    make_dir(DATA_PATH + "Heatmaps/GR")
    make_dir(DATA_PATH + "Heatmaps/Ymax")
    
    # Checks to see if plate mapping file exists is present
    map_file = ""
    try:
        for file in os.listdir(DATA_PATH):
            if file == "plate_mapping.txt" or file == "plate_mappings.txt":
                map_file = file
                # Pulling info from experimental_info.txt
                print("Using information from plate_mapping.txt\n")
                break
    except FileNotFoundError:
        print("FileNotFoundError, try making a plate_mapping.txt file and trying again")
        exit(0)
    if map_file == "":
        print("FileNotFoundError, try making a plate_mapping.txt file and trying again")
        exit(0)
        
    # Build data structures for comparisons
    plate_names = list(df_dict.keys())
    # Initialize plate objects for each plate
    plate_obj_list = {}
    for plate_name in plate_names:
        plate_obj_list[plate_name] = Plate(plate_name) 
        
    # List to hold all comparisons being made
    comparisons = []
    # Set to hold all replicate names
    replicates = set()
    # Parse plaate_mapping.txt for instructions
    with open(DATA_PATH + map_file, "r") as map_f:
        for line in map_f:
            # Mapping of experimental replicates
            if "replicates" in line.lower():
                line = map_f.readline()
                while line != "\n" and line != "":
                    # Plate name
                    if line.strip(":\n") in plate_names:
                        # Create a plate object
                        plate_name = line.strip(":\n")                        
                        line = map_f.readline()
                        # Replicate mapping in plate
                        while line != "\n" and line != "" and line.replace(":\n", "") not in plate_names:
                            wells = []
                            tokens = line.split(":")
                            repl_name = tokens[0]
                            replicates.add(repl_name)
                            well_str = tokens[1]
                            indiv_wells = well_str.split(",")
                            for well in indiv_wells:
                                wells.append(well.strip(", \n"))
                            # Add replicate info to the corresponding plate object
                            plate_obj_list[plate_name].add_repl(repl_name, wells)
                            line = map_f.readline()
            
            elif "compare" in line.lower():
                line = map_f.readline()
                while line != "\n" and line != "":
                    comp = []
                    tokens = line.split(":")
                    for token in tokens:
                        comp.append(token.strip(" \n"))
                    comparisons.append(comp)
                    line = map_f.readline()

    for comparison in comparisons:
        # Comparing two strains
        if len(comparison) == 2:
            exp_repl = comparison[0]
            exp_data = {}
            con_repl = comparison[1]
            con_data = {}
            # Loop through plates pulling out groupings of replicates
            for plate_name in plate_obj_list.keys():
                plat = plate_obj_list[plate_name]
                if exp_repl in plat.get_repl_names():
                    exp_data[plate_name] = plat.get_wells(exp_repl)
                    
                if con_repl in plat.get_repl_names():
                    con_data[plate_name] = plat.get_wells(con_repl)
                    
            # If log flag was chosen create extra directory
            if log_flag:
                make_dir(DATA_PATH + "Graphs/" + exp_repl + "/Log2_Graphs")
                
        # Comparing three strains comping in future update
        elif len(comparison) == 3:
            pass
        
        # Graph averages of replicates against each other
        graph_avg(df_dict, con_data, exp_data, con_repl, exp_repl, DATA_PATH, plate_obj_list, hm_flag, log_flag)
        # Grapah individual wells of replicate
        if exp_repl in replicates:
            graph_indiv(df_dict,  exp_data, exp_repl, DATA_PATH, plate_obj_list, log_flag)
            replicates.remove(exp_repl)
        if con_repl in replicates:
            graph_indiv(df_dict, con_data, con_repl, DATA_PATH, plate_obj_list, log_flag)
            replicates.remove(con_repl)
    

    if replicates:
        replicates = list(replicates)
        for repl in replicates:
            repl_data = {}
            for plate_name in plate_obj_list.keys():
                plat = plate_obj_list[plate_name]
                if repl in plat.get_repl_names():
                    repl_data[plate_name] = plat.get_wells(repl)
            # If log flag was chosen create extra directory
            if log_flag:
                make_dir(DATA_PATH + "Graphs/" + repl + "/Log2_Graphs")
                
            graph_indiv(df_dict, repl_data, repl, DATA_PATH, plate_obj_list, log_flag)
            replicates.remove(repl)
            
    print("Heatmaps...")
    # Loop through plates to make heatmaps
    for plate_name in plate_obj_list.keys():
        plat = plate_obj_list[plate_name]
        hm(plat, DATA_PATH)
        #heatmap_ymax()

    print("\nFinished.")
    
""" Auxillary Functions """
def make_dir(dir_name):
    try:
        os.mkdir(dir_name)
    except OSError:
       #print("Creation of the directory %s failed" % (dir_name))
       pass
       
if __name__ == "__main__":
    proces(DATA_PATH_HC, hm_flag = False, log_flag = False)